@@ -1,8 +1,7 @@
 /*
  *      libmain.c - this file is part of Geany, a fast and lightweight IDE
  *
- *      Copyright 2005-2012 Enrico Tröger <enrico(dot)troeger(at)uvena(dot)de>
- *      Copyright 2006-2012 Nick Treleaven <nick(dot)treleaven(at)btinternet(dot)com>
+ *      Copyright 2005 The Geany contributors
  *
  *      This program is free software; you can redistribute it and/or modify
  *      it under the terms of the GNU General Public License as published by
@@ -64,8 +63,6 @@
 #include "win32.h"
 #include "osx.h"
 
-#include "gtkcompat.h"
-
 #include <signal.h>
 #include <time.h>
 #include <sys/types.h>
@@ -74,8 +71,13 @@
 #include <string.h>
 #include <stdlib.h>
 
+#include <gtk/gtk.h>
 #include <glib/gstdio.h>
 
+#ifdef G_OS_UNIX
+# include <glib-unix.h>
+#endif
+
 #ifdef HAVE_LOCALE_H
 # include <locale.h>
 #endif
@@ -115,17 +117,17 @@ static gboolean dummy = FALSE;
 /* in alphabetical order of short options */
 static GOptionEntry entries[] =
 {
-	{ "column", 0, 0, G_OPTION_ARG_INT, &cl_options.goto_column, N_("Set initial column number for the first opened file (useful in conjunction with --line)"), NULL },
-	{ "config", 'c', 0, G_OPTION_ARG_FILENAME, &alternate_config, N_("Use an alternate configuration directory"), NULL },
+	{ "column", 0, 0, G_OPTION_ARG_INT, &cl_options.goto_column, N_("Set initial column number to COLUMN for the first opened file (useful in conjunction with --line)"), N_("COLUMN") },
+	{ "config", 'c', 0, G_OPTION_ARG_FILENAME, &alternate_config, N_("Use alternate configuration directory DIR"), N_("DIR") },
 	{ "ft-names", 0, 0, G_OPTION_ARG_NONE, &ft_names, N_("Print internal filetype names"), NULL },
 	{ "generate-tags", 'g', 0, G_OPTION_ARG_NONE, &generate_tags, N_("Generate global tags file (see documentation)"), NULL },
 	{ "no-preprocessing", 'P', 0, G_OPTION_ARG_NONE, &no_preprocessing, N_("Don't preprocess C/C++ files when generating tags file"), NULL },
 #ifdef HAVE_SOCKET
 	{ "new-instance", 'i', 0, G_OPTION_ARG_NONE, &cl_options.new_instance, N_("Don't open files in a running instance, force opening a new instance"), NULL },
-	{ "socket-file", 0, 0, G_OPTION_ARG_FILENAME, &cl_options.socket_filename, N_("Use this socket filename for communication with a running Geany instance"), NULL },
+	{ "socket-file", 0, 0, G_OPTION_ARG_FILENAME, &cl_options.socket_filename, N_("Use socket filename FILE for communication with a running Geany instance"), N_("FILE") },
 	{ "list-documents", 0, 0, G_OPTION_ARG_NONE, &cl_options.list_documents, N_("Return a list of open documents in a running Geany instance"), NULL },
 #endif
-	{ "line", 'l', 0, G_OPTION_ARG_INT, &cl_options.goto_line, N_("Set initial line number for the first opened file"), NULL },
+	{ "line", 'l', 0, G_OPTION_ARG_INT, &cl_options.goto_line, N_("Set initial line number to LINE for the first opened file"), N_("LINE") },
 	{ "no-msgwin", 'm', 0, G_OPTION_ARG_NONE, &no_msgwin, N_("Don't show message window at startup"), NULL },
 	{ "no-ctags", 'n', 0, G_OPTION_ARG_NONE, &ignore_global_tags, N_("Don't load auto completion data (see documentation)"), NULL },
 #ifdef HAVE_PLUGINS
@@ -136,7 +138,7 @@ static GOptionEntry entries[] =
 	{ "no-session", 's', G_OPTION_FLAG_REVERSE, G_OPTION_ARG_NONE, &cl_options.load_session, N_("Don't load the previous session's files"), NULL },
 #ifdef HAVE_VTE
 	{ "no-terminal", 't', 0, G_OPTION_ARG_NONE, &no_vte, N_("Don't load terminal support"), NULL },
-	{ "vte-lib", 0, 0, G_OPTION_ARG_FILENAME, &lib_vte, N_("Filename of libvte.so"), NULL },
+	{ "vte-lib", 0, 0, G_OPTION_ARG_FILENAME, &lib_vte, N_("Use FILE as the dynamically-linked VTE library"), N_("FILE") },
 #endif
 	{ "verbose", 'v', 0, G_OPTION_ARG_NONE, &verbose_mode, N_("Be verbose"), NULL },
 	{ "version", 'V', 0, G_OPTION_ARG_NONE, &show_version, N_("Show version and exit"), NULL },
@@ -148,19 +150,27 @@ static GOptionEntry entries[] =
 static void setup_window_position(void)
 {
 	/* interprets the saved window geometry */
-	if (!prefs.save_winpos)
-		return;
+	if (prefs.save_wingeom)
+	{
+		if (ui_prefs.geometry[2] != -1 && ui_prefs.geometry[3] != -1)
+		{
+			gtk_window_set_default_size(GTK_WINDOW(main_widgets.window),
+				ui_prefs.geometry[2], ui_prefs.geometry[3]);
+		}
+	}
 
-	if (ui_prefs.geometry[0] != -1 && ui_prefs.geometry[1] != -1)
-		gtk_window_move(GTK_WINDOW(main_widgets.window),
-			ui_prefs.geometry[0], ui_prefs.geometry[1]);
-
-	if (ui_prefs.geometry[2] != -1 && ui_prefs.geometry[3] != -1)
-		gtk_window_set_default_size(GTK_WINDOW(main_widgets.window),
-			ui_prefs.geometry[2], ui_prefs.geometry[3]);
-
-	if (ui_prefs.geometry[4] == 1)
-		gtk_window_maximize(GTK_WINDOW(main_widgets.window));
+	if (prefs.save_winpos)
+	{
+		if (ui_prefs.geometry[0] != -1 && ui_prefs.geometry[1] != -1)
+		{
+			gtk_window_move(GTK_WINDOW(main_widgets.window),
+				ui_prefs.geometry[0], ui_prefs.geometry[1]);
+		}
+		if (ui_prefs.geometry[4] == 1)
+		{
+			gtk_window_maximize(GTK_WINDOW(main_widgets.window));
+		}
+	}
 }
 
 
@@ -279,7 +289,7 @@ static void main_init(void)
 	osx_ui_init();
 #endif
 
-	/* set widget names for matching with .gtkrc-2.0 */
+	/* set widget names for matching with GTK CSS */
 	gtk_widget_set_name(main_widgets.window, "GeanyMainWindow");
 	gtk_widget_set_name(ui_widgets.toolbar_menu, "GeanyToolbarMenu");
 	gtk_widget_set_name(main_widgets.editor_menu, "GeanyEditMenu");
@@ -384,6 +394,26 @@ static void get_line_and_column_from_filename(gchar *filename, gint *line, gint
 
 
 #ifdef G_OS_WIN32
+static gint get_windows_socket_port(void)
+{
+	/* Read config file early to get TCP port number as we need it for IPC before all
+	 * other settings are read in load_settings() */
+	gchar *configfile = g_build_filename(app->configdir, "geany.conf", NULL);
+	GKeyFile *config = g_key_file_new();
+	gint port_number;
+
+	g_key_file_load_from_file(config, configfile, G_KEY_FILE_NONE, NULL);
+	port_number = utils_get_setting_integer(config, PACKAGE, "socket_remote_cmd_port",
+		SOCKET_WINDOWS_REMOTE_CMD_PORT);
+	geany_debug("Using TCP port number %d for IPC", port_number);
+	g_free(configfile);
+	g_key_file_free(config);
+	g_return_val_if_fail(port_number >= 1024 && port_number <= (gint)G_MAXUINT16,
+		SOCKET_WINDOWS_REMOTE_CMD_PORT);
+	return port_number;
+}
+
+
 static void change_working_directory_on_windows(void)
 {
 	gchar *install_dir = win32_get_installation_dir();
@@ -523,6 +553,8 @@ static void parse_command_line_options(gint *argc, gchar ***argv)
 
 	context = g_option_context_new(_("[FILES...]"));
 
+	g_option_context_set_summary(context, _("A fast and lightweight IDE."));
+	g_option_context_set_description(context, _("Report bugs to https://github.com/geany/geany/issues."));
 	g_option_context_add_main_entries(context, entries, GETTEXT_PACKAGE);
 	g_option_group_set_translation_domain(g_option_context_get_main_group(context), GETTEXT_PACKAGE);
 	g_option_context_add_group(context, gtk_get_option_group(FALSE));
@@ -760,21 +792,25 @@ static gint setup_config_dir(void)
 	}
 	/* make configdir a real path */
 	if (g_file_test(app->configdir, G_FILE_TEST_EXISTS))
-		SETPTR(app->configdir, tm_get_real_path(app->configdir));
+		SETPTR(app->configdir, utils_get_real_path(app->configdir));
 
 	return mkdir_result;
 }
 
-/* Signal handling removed since main_quit() uses functions that are
- * illegal in signal handlers
-static void signal_cb(gint sig)
+
+#ifdef G_OS_UNIX
+static gboolean signal_cb(gpointer user_data)
 {
+	gint sig = GPOINTER_TO_INT(user_data);
 	if (sig == SIGTERM)
 	{
+		geany_debug("Received SIGTERM signal");
 		main_quit();
 	}
+	return G_SOURCE_REMOVE;
 }
- */
+#endif
+
 
 /* Used for command-line arguments at startup or from socket.
  * this will strip any :line:col filename suffix from locale_filename */
@@ -970,52 +1006,6 @@ static const gchar *get_locale(void)
 }
 
 
-#if ! GTK_CHECK_VERSION(3, 0, 0)
-/* This prepends our own gtkrc file to the list of RC files to be loaded by GTK at startup.
- * This function *has* to be called before gtk_init().
- *
- * We have a custom RC file defining various styles we need, and we want the user to be
- * able to override them (e.g. if they want -- or need -- other colors).  Fair enough, one
- * would simply call gtk_rc_parse() with the appropriate filename.  However, the styling
- * rules applies in the order they are loaded, then if we load our styles after GTK has
- * loaded the user's ones we'd override them.
- *
- * There are 2 solutions to fix this:
- * 1) set our styles' priority to something with lower than "user" (actually "theme"
- *    priority because rules precedence are first calculated depending on the priority
- *    no matter of how precise the rules is, so we need to override the theme).
- * 2) prepend our custom style to GTK's list while keeping priority to user (which is the
- *    default), so it gets loaded before real user's ones and so gets overridden by them.
- *
- * One would normally go for 1 because it's ways simpler and requires less code: you just
- * have to add the priorities to your styles, which is a matter of adding a few ":theme" in
- * the RC file.  However, KDE being a bitch it doesn't set the gtk-theme-name but rather
- * directly includes the style to use in a user gtkrc file, which makes the theme have
- * "user" priority, hence overriding our styles.  So, we cannot set priorities in the RC
- * file if we want to support running under KDE, which pretty much leave us with no choice
- * but to go with solution 2, which unfortunately requires writing ugly code since GTK
- * don't have a gtk_rc_prepend_default_file() function.  Thank you very much KDE.
- *
- * Though, as a side benefit it also makes the code work with people using gtk-chtheme,
- * which also found it funny to include the theme in the user RC file. */
-static void setup_gtk2_styles(void)
-{
-	gchar **gtk_files = gtk_rc_get_default_files();
-	gchar **new_files = g_malloc(sizeof *new_files * (g_strv_length(gtk_files) + 2));
-	guint i = 0;
-
-	new_files[i++] = g_build_filename(app->datadir, "geany.gtkrc", NULL);
-	for (; *gtk_files; gtk_files++)
-		new_files[i++] = g_strdup(*gtk_files);
-	new_files[i] = NULL;
-
-	gtk_rc_set_default_files(new_files);
-
-	g_strfreev(new_files);
-}
-#endif
-
-
 GEANY_EXPORT_SYMBOL
 gint main_lib(gint argc, gchar **argv)
 {
@@ -1023,6 +1013,7 @@ gint main_lib(gint argc, gchar **argv)
 	gint config_dir_result;
 	const gchar *locale;
 	gchar *utf8_configdir;
+	gchar *os_info;
 
 #if ! GLIB_CHECK_VERSION(2, 36, 0)
 	g_type_init();
@@ -1043,9 +1034,7 @@ gint main_lib(gint argc, gchar **argv)
 	memset(&ui_widgets, 0, sizeof(UIWidgets));
 
 	setup_paths();
-#if ! GTK_CHECK_VERSION(3, 0, 0)
-	setup_gtk2_styles();
-#endif
+
 #ifdef ENABLE_NLS
 	main_locale_init(utils_resource_dir(RESOURCE_DIR_LOCALE), GETTEXT_PACKAGE);
 #endif
@@ -1060,10 +1049,9 @@ gint main_lib(gint argc, gchar **argv)
 		g_thread_init(NULL);
 #endif
 
-	/* removed as signal handling was wrong, see signal_cb()
-	signal(SIGTERM, signal_cb); */
-
 #ifdef G_OS_UNIX
+	g_unix_signal_add(SIGTERM, signal_cb, GINT_TO_POINTER(SIGTERM));
+
 	/* ignore SIGPIPE signal for preventing sudden death of program */
 	signal(SIGPIPE, SIG_IGN);
 #endif
@@ -1073,9 +1061,13 @@ gint main_lib(gint argc, gchar **argv)
 	/* check and create (unix domain) socket for remote operation */
 	if (! socket_info.ignore_socket)
 	{
+		gushort socket_port = 0;
+#ifdef G_OS_WIN32
+		socket_port = (gushort) get_windows_socket_port();
+#endif
 		socket_info.lock_socket = -1;
 		socket_info.lock_socket_tag = 0;
-		socket_info.lock_socket = socket_init(argc, argv);
+		socket_info.lock_socket = socket_init(argc, argv, socket_port);
 		/* Quit if filenames were sent to first instance or the list of open
 		 * documents has been printed */
 		if ((socket_info.lock_socket == -2 /* socket exists */ && argc > 1) ||
@@ -1112,6 +1104,14 @@ gint main_lib(gint argc, gchar **argv)
 	geany_debug(geany_lib_versions,
 		gtk_major_version, gtk_minor_version, gtk_micro_version,
 		glib_major_version, glib_minor_version, glib_micro_version);
+
+	os_info = utils_get_os_info_string();
+	if (os_info != NULL)
+	{
+		geany_debug("OS: %s", os_info);
+		g_free(os_info);
+	}
+
 	geany_debug("System data dir: %s", app->datadir);
 	utf8_configdir = utils_get_utf8_from_locale(app->configdir);
 	geany_debug("User config dir: %s", utf8_configdir);
@@ -1157,8 +1157,20 @@ gint main_lib(gint argc, gchar **argv)
 
 	ui_set_statusbar(TRUE, _("This is Geany %s."), main_get_version_string());
 	if (config_dir_result != 0)
-		ui_set_statusbar(TRUE, _("Configuration directory could not be created (%s)."),
-			g_strerror(config_dir_result));
+	{
+		const gchar *message = _("Configuration directory could not be created (%s).");
+		ui_set_statusbar(TRUE, message, g_strerror(config_dir_result));
+		g_warning(message, g_strerror(config_dir_result));
+	}
+#ifdef HAVE_SOCKET
+	if (socket_info.lock_socket == -1)
+	{
+		const gchar *message =
+			_("IPC socket could not be created, see Help->Debug Messages for details.");
+		ui_set_statusbar(TRUE, "%s", message);
+		g_warning("%s", message);
+	}
+#endif
 
 	/* apply all configuration options */
 	apply_settings();
@@ -1197,11 +1209,6 @@ gint main_lib(gint argc, gchar **argv)
 	build_menu_update(doc);
 	sidebar_update_tag_list(doc, FALSE);
 
-#ifdef G_OS_WIN32
-	/* Manually realise the main window to be able to set the position but don't show it.
-	 * We don't set the position after showing the window to avoid flickering. */
-	gtk_widget_realize(main_widgets.window);
-#endif
 	setup_window_position();
 
 	/* finally show the window */
@@ -1245,16 +1252,20 @@ static void queue_free(GQueue *queue)
 }
 
 
-static void do_main_quit(void)
+static gboolean do_main_quit(void)
 {
-	geany_debug("Quitting...");
-
 	configuration_save();
 
 	if (app->project != NULL)
-		project_close(FALSE);   /* save project session files */
+	{
+		if (!project_close(FALSE))   /* save project session files */
+			return FALSE;
+	}
 
-	document_close_all();
+	if (!document_close_all())
+		return FALSE;
+
+	geany_debug("Quitting...");
 
 	main_status.quitting = TRUE;
 
@@ -1293,6 +1304,7 @@ static void do_main_quit(void)
 	g_free(prefs.default_open_path);
 	g_free(prefs.custom_plugin_path);
 	g_free(ui_prefs.custom_date_format);
+	g_free(ui_prefs.color_picker_palette);
 	g_free(interface_prefs.editor_font);
 	g_free(interface_prefs.tagbar_font);
 	g_free(interface_prefs.msgwin_font);
@@ -1347,6 +1359,8 @@ static void do_main_quit(void)
 	ui_finalize_builder();
 
 	gtk_main_quit();
+
+	return TRUE;
 }
 
 
@@ -1372,19 +1386,16 @@ gboolean main_quit(void)
 
 	if (! check_no_unsaved())
 	{
-		if (document_account_for_unsaved())
-		{
-			do_main_quit();
+		if (do_main_quit())
 			return TRUE;
-		}
 	}
 	else
 	if (! prefs.confirm_exit ||
 		dialogs_show_question_full(NULL, GTK_STOCK_QUIT, GTK_STOCK_CANCEL, NULL,
 			_("Do you really want to quit?")))
 	{
-		do_main_quit();
-		return TRUE;
+		if (do_main_quit())
+			return TRUE;
 	}
 
 	main_status.quitting = FALSE;
