@@ -1,8 +1,7 @@
 /*
  *      pluginutils.c - this file is part of Geany, a fast and lightweight IDE
  *
- *      Copyright 2009-2012 Nick Treleaven <nick(dot)treleaven(at)btinternet(dot)com>
- *      Copyright 2009-2012 Enrico Tröger <enrico(dot)troeger(at)uvena(dot)de>
+ *      Copyright 2005 The Geany contributors
  *
  *      This program is free software; you can redistribute it and/or modify
  *      it under the terms of the GNU General Public License as published by
@@ -396,7 +395,7 @@ static GtkWidget *create_pref_page(Plugin *p, GtkWidget *dialog)
 
 			gtk_alignment_set_padding(GTK_ALIGNMENT(align), 6, 6, 6, 6);
 			gtk_container_add(GTK_CONTAINER(align), page);
-			page = gtk_vbox_new(FALSE, 0);
+			page = gtk_box_new(GTK_ORIENTATION_VERTICAL, 0);
 			gtk_box_pack_start(GTK_BOX(page), align, TRUE, TRUE, 0);
 		}
 	}
@@ -570,6 +569,24 @@ void plugin_builder_connect_signals(GeanyPlugin *plugin,
 }
 
 
+/** Get the additional data that corresponds to the plugin.
+ *
+ * @param plugin The plugin provided by Geany
+ * @return The data corresponding to the plugin or @c NULL if none set.
+ *
+ * @since 1.32 (API 234)
+ *
+ * @see geany_plugin_set_data()
+ */
+gpointer geany_plugin_get_data(const GeanyPlugin *plugin)
+{
+	g_return_val_if_fail (plugin != NULL, NULL);
+	g_return_val_if_fail (PLUGIN_LOADED_OK (plugin->priv), NULL);
+
+	return plugin->priv->cb_data;
+}
+
+
 /** Add additional data that corresponds to the plugin.
  *
  * @p pdata is the pointer going to be passed to the individual plugin callbacks
@@ -706,7 +723,7 @@ void plugin_set_document_data(struct GeanyPlugin *plugin, struct GeanyDocument *
  * void on_document_save(GObject *unused, GeanyDocument *doc, GeanyPlugin *plugin)
  * {
  *     const gchar *some_data = plugin_get_document_data(plugin, doc, "my-data");
- *     g_print("my-data: %s", some_data);
+ *     g_print("my-data: %s\n", some_data);
  * }
  *
  * gboolean plugin_init(GeanyPlugin *plugin, gpointer unused)
